import functools
import os
from pathlib import Path
from re import S
from typing import NotRequired, TypedDict
from yaml import safe_load

from packe._config import ConfigFile


class ConfigFileWrapper:
    _data: ConfigFile

    def __init__(self, path: Path):
        self.path = path
        self._data = safe_load(path.read_text())
        before_str = self._data.get("before", None)
        self.before = (
            path.absolute().parent / Path(before_str) if before_str else None
        )
        self.root_only = self._data.get("root_only", False)

    @functools.cached_property
    def root_pack(self):
        from packe._scripts import Pack

        root = Pack.root()
        for name, obj in self._data["entrypoints"].items():
            path = self.path.absolute().parent / obj["path"]
            expanded = os.path.expandvars(path)
            root.add(Pack.from_indexed_dir(root, Path(expanded), name))
        return root



from dataclasses import dataclass
from pathlib import Path
from packe._exec.bash_exec_prefix import BashPrefixExecutor
from packe._scripts.indexed import (
    is_valid_script,
    must_parse_indexed,
    try_parse_indexed,
)
from packe._scripts.pretty_print import pretty_print_lines
from packe._scripts.runnable import Runnable


@dataclass(eq=False)
class Script(Runnable):
    path: Path
    name: str

    def __eq__(self, value: object) -> bool:
        return super().__eq__(value)

    def __hash__(self) -> int:
        return super().__hash__()

    @staticmethod
    def is_valid_indexed(path: Path):
        return try_parse_indexed(path.stem) is not None and is_valid_script(
            path
        )

    @classmethod
    def from_indexed_path(cls, parent: Runnable, path: Path):
        if not path.is_file():
            raise Exception(f"Expected a file, got {path}")
        if not cls.is_valid_indexed(path):
            raise Exception(f"Invalid indexed path: {path}")
        [pos, str] = must_parse_indexed(path.stem)
        return Script(pos, str, parent, path)

    @staticmethod
    def from_named(
        parent: Runnable | None, path: Path, name: str | None = None
    ):
        name = name or path.stem
        return Script(None, name, parent, path)

    def __repr__(self) -> str:
        return f"{self:line}"

    @property
    def contents(self):
        return self.path.read_text()

    def __len__(self):
        return len(self.contents)

    def __format__(self, format_spec: str = "short") -> str:
        format_spec = format_spec or "short"
        match format_spec:
            case "full":
                lines = pretty_print_lines(self.address, self.contents)
                return lines
            case "child":
                return f"{self.name}[{len(self)}]!"
            case "line":
                return f"{self.address}[{len(self)}]!"
            case "short":
                return f"{self.caption}!"
            case "address":
                return f"{self.address}!"
            case _:
                raise ValueError(f"Unknown format spec: {format_spec}")

    def __str__(self) -> str:
        if self.pos is not None:
            return f"{self.pos:02d}:{self.name}"
        return self.name

    def run(self, executor: BashPrefixExecutor):
        cwd = self.path.parent
        executor.must_exec(self.path, cwd, self.address)
        # code to exec



import importlib.metadata
import os
import sys
from pathlib import Path
from sys import platform

from termcolor import colored

from packe._cli import _Cli
from packe._config_wrapper import ConfigFileWrapper
from packe._exec.bash_exec_prefix import BashPrefixExecutor
from packe._print import fatal_error


def start():
    if platform == "win32" or platform == "darwin":
        fatal_error("packe only supports Linux!", 3)

    cli = _Cli()
    command = cli.parse()
    if command.command == "version":
        v = importlib.metadata.version("packe")
        print(colored(f"packe {v} / python {sys.version}", "green"))
        exit(0)
    cfg = ConfigFileWrapper(Path(command.config.strip()))
    if os.geteuid() != 0 and cfg.root_only:
        fatal_error("Config file requires running as root!", 4)

    pack = cfg.root_pack
    if len(pack) == 0:
        fatal_error("No scripts found in config", 1)

    matched = pack.find_all(command.selector)
    executor = BashPrefixExecutor(cfg.before)
    cmd = command.command
    if hasattr(command, "dry") and command.dry:
        cmd = "list"
    if not matched:
        splat = " ".join(command.selector)
        fatal_error(f"No scripts found for {splat}", 10)
    print(f"Matched {len(matched)} scripts")
    match cmd:
        case "run":
            matched.run(executor)
        case "list":
            print(f"{matched:summary}")
        case "print":
            print(f"{matched:full}")
        case _:
            fatal_error(f"Unknown command: {cmd}", 2)
    exit(0)



from __future__ import annotations


from typing import TYPE_CHECKING, Iterable, Callable, TypeAlias, Union


if TYPE_CHECKING:
    from packe._scripts.runnable import Runnable

    Selector: TypeAlias = Callable[[Runnable], bool]


def indexed_selector() -> Callable[[Runnable], bool]:

    def rule(script: Runnable):
        return script.pos is not None

    return rule


def range_selector(
    start: int | None, end: int | None
) -> Callable[[Runnable], bool]:

    def rule(script: Runnable):
        if script.pos is None:
            return False
        if start is not None and script.pos < start:
            return False
        if end is not None and script.pos > end:
            return False
        return True

    return rule


def pos_selector(pos: int) -> Callable[[Runnable], bool]:

    def rule(script: Runnable):
        return script.pos == pos

    return rule


def name_selector(name: str) -> Callable[[Runnable], bool]:

    def rule(script: Runnable):
        return script.name and script.name.lower() == name.lower() or False

    return rule


def any_selector(
    selectors: Iterable[Callable[[Runnable], bool]],
) -> Callable[[Runnable], bool]:

    def rule(script: Runnable):
        return bool([1 for rule in selectors if rule(script)])

    return rule


def parse_selector(selector: str) -> Callable[[Runnable], bool]:
    if selector == "%":
        return indexed_selector(True)
    elif "-" in selector:
        [before, after] = selector.split("-")
        # Either '-1' or '1-2':
        if not before or before.isdigit():
            return range_selector(
                int(before) if before else None, int(after) if after else None
            )
        else:
            # Must be a name with a '-' in it.
            return name_selector(selector)
    elif selector.isdigit():
        return pos_selector(int(selector))
    else:
        return name_selector(selector)


def parse_selector_list(union_selector_str: str) -> Callable[[Runnable], bool]:
    selector_list = union_selector_str.split(",")
    return any_selector([parse_selector(s) for s in selector_list])



from typing import NotRequired, TypedDict


class ConfigEntry(TypedDict):
    path: str


class ConfigFile(TypedDict):
    root_only: NotRequired[bool]
    entrypoints: dict[str, ConfigEntry]
    prerun: NotRequired[str]



from time import sleep
from termcolor import colored
from packe._exec.bash_exec_prefix import BashPrefixExecutor
from packe._scripts.runnable import Runnable
from packe._scripts.script import Script


class MatchedSet(Runnable):

    def __init__(self, multipart: list[str], kids: list[Runnable]):
        super().__init__(
            pos=None,
            name=None,
            parent=None,
        )
        self.query = ", ".join(multipart)
        self.kids = kids

    def run(self, executor: BashPrefixExecutor):
        from .pack import Pack

        failed_parents_set = {
            x
            for y in self.kids
            for x in [y, *y.parents]
            if isinstance(x, Pack) and x.pre_run
            if not x.do_prerun(executor)
        }

        for kid in self.kids:
            parents = {*kid.parents, kid}
            intersection = parents.intersection(failed_parents_set)
            if intersection:
                first_failed_parent = next(iter(intersection), None)
                print(
                    colored(
                        f"║ '{str(kid)}' SKIPPED; FAILED PRERUN '{first_failed_parent}' ║",
                        color="black",
                        on_color="on_red",
                    )
                )
                continue

            kid.run(executor)
        sleep(0.1)
        print(
            colored(
                f"      ↑ DONE ↑      ",
                color="white",
                on_color="on_light_green",
            ),
            "\n",
        )

    def __len__(self) -> int:
        from packe._scripts.pack import Pack

        count = 0
        for x in self.kids:
            match x:
                case Pack():
                    count += len(x)
                case Script():
                    count += 1
                case _:
                    raise ValueError(f"Unknown type: {type(x)}")
        return count

    def __iter__(self):
        return iter(self.kids)

    def __format__(self, format_spec: str) -> str:
        format_spec = format_spec or "short"
        match format_spec:
            case "full":
                texts = []
                for kid in self.kids:
                    texts.append(f"{kid:full}")
                return "\n".join(texts)
            case "summary":
                lines = []
                for script in self:
                    lines.append(f"{script:short}")
                return "\n".join(lines)
            case "child":
                return f"{self.query}[{len(self)}]!"
            case "line":
                return f"{self.query}[{len(self)}]!"
            case "short":
                return f"{self.query}!"
            case "address":
                return f"{self.query}!"
            case _:
                raise ValueError(f"Unknown format spec: {format_spec}")

    def __repr__(self) -> str:
        return self.__format__("line")

    def __str__(self) -> str:
        return self.__format__("short")



import argparse
from logging import fatal
from os import environ
from sys import stderr
from typing import Iterable, Sequence, Union
from ._print import fatal_error

from packe._command import Command
from packe._matching.script_selectors import parse_selector_list


def _add_selector(p: argparse.ArgumentParser):
    p.add_argument(
        "selector",
        nargs="+",
        help="one or more run selectors for packe scripts and packs to run.",
    )


class _Cli:
    def __init__(self):
        root_parser = argparse.ArgumentParser(
            description="Perdido setup script runner"
        )

        root_parser.add_argument(
            "-C",
            "--config",
            help="path to a config file",
            required=False,
            type=str,
        )
        subparsers = root_parser.add_subparsers(
            title="command", required=True, dest="command"
        )

        version = subparsers.add_parser(
            "version", help="print the version and exit"
        )

        run = subparsers.add_parser(
            "run",
            help="run one or more installation scripts",
        )

        run.add_argument(
            "-D",
            "--dry",
            dest="dry",
            action="store_true",
            help="don't actually run the scripts",
            required=False,
        )
        _add_selector(run)
        list = subparsers.add_parser("list", help="list installation scripts")
        _add_selector(list)
        printing = subparsers.add_parser(
            "print", help="print installation scripts"
        )
        _add_selector(printing)

        self._parser = root_parser

    def parse(self, args: Union[Sequence[str], None] = None) -> Command:
        args_result = self._parser.parse_args(args)
        if args_result.command == "version":
            return args_result  # type: ignore
        if hasattr(args_result, "rule"):
            args_result.rule = parse_selector_list(args_result.rule)
        if not hasattr(args_result, "config"):
            args_result.config = environ.get(
                "PYRUN_CONFIG", None
            ) or environ.get("PACKE_CONFIG", None)
            if not args_result.config:
                fatal_error(
                    "No config file specified and PACKE_CONFIG not set", 2
                )

        return args_result  # type: ignore



from typing import Iterable

from termcolor import colored
from packe._scripts.runnable import Runnable
from pygments import highlight
from pygments.formatters.terminal import TerminalFormatter
from pygments.lexers.shell import BashLexer

term = TerminalFormatter()
lexer = BashLexer()


def pretty_print_lines(title: str, contents: str):
    # NNN | Line
    title = colored(title, attrs=["underline"])
    contents = contents.strip()
    contents = highlight(contents, lexer, term)
    lines = contents.split("\n")
    lines = [f"{(i + 1):>3} | {line}" for i, line in enumerate(lines)]
    return "\n".join([title, *lines])


def pretty_print_kids(title: str, kids: Iterable["Runnable"]):
    results = [f"∙ {x.name}" for x in kids]
    title = colored(title, attrs=["underline"])
    return "\n".join([title, *results])



import sys
from typing import NoReturn

from termcolor import colored


def echo_warn(message: str):
    message = colored(f"⚠️ {message}", "yellow")
    print(message, file=sys.stdout)


def fatal_error(message: str, code: int = 1) -> NoReturn:
    message = colored(f"💀 {message}", "black", "on_red")
    print(message, file=sys.stderr)
    exit(code)



from calendar import c
from dataclasses import field, dataclass
from os import name
from pathlib import Path
from typing import Callable, Iterable

from termcolor import colored
from packe._exec.bash_exec_prefix import BashPrefixExecutor
from packe._matching.script_selectors import parse_selector_list
from packe._scripts.matched_set import MatchedSet
from packe._scripts.pretty_print import pretty_print_kids
from packe._scripts.script import Script
from packe._scripts.indexed import must_parse_indexed, try_parse_indexed
from packe._scripts.runnable import Runnable
from packe._scripts.types import RunnableFormat


@dataclass(eq=False)
class Pack(Runnable):
    name: str
    pre_run: Script | None = field(default=None)

    _kids: list[Runnable] = field(init=False, default_factory=list)

    @property
    def kids(self):
        return [x for x in sorted(self._kids, key=lambda x: x.pos or 0)]

    @classmethod
    def is_valid_indexed(cls, path: Path):
        return path.is_dir() and try_parse_indexed(path.name) is not None

    @classmethod
    def root(cls):
        pk = Pack(None, "@", None)
        return pk

    def add(self, *kids: Runnable):
        self._kids = [*self._kids, *kids]
        return self

    @classmethod
    def from_indexed_dir(
        cls, parent: Runnable | None, index_root: Path, name: str | None = None
    ):
        if not index_root.is_dir():
            raise Exception(
                f"Expected a directory, got {index_root}, ${index_root.stat()}"
            )
        if name is not None:
            pos = None
        elif cls.is_valid_indexed(index_root):
            pos, name = must_parse_indexed(index_root.name)
        else:
            name = index_root.name
            pos = None
        prerun_path = [
            *index_root.glob("pyrun.pre*"),
            *index_root.glob("packe.pre*"),
        ]
        prerun_script: Script | None = None
        if prerun_path:
            prerun_path = prerun_path[0]
            prerun_script = Script.from_named(parent, prerun_path)
        p = Pack(pos, name, parent, prerun_script)

        def make_kid(path: Path):
            if path.is_dir():
                return Pack.from_indexed_dir(p, path)
            assert parent is not None
            return Script.from_indexed_path(p, path)

        indexed_kids = [
            make_kid(x)
            for x in index_root.glob("*")
            if try_parse_indexed(x.name) is not None
        ]
        p.add(*indexed_kids)
        return p

    def __bool__(self):
        return bool(self.kids)

    def find_all(self, multipart_selectors: list[str]):
        all_results: list[Runnable] = []
        for multipart_selector in multipart_selectors:
            selector_list = [
                parse_selector_list(s) for s in multipart_selector.split("/")
            ]

            results = [self]
            for selector in selector_list:
                next_results = []
                for x in results:
                    assert isinstance(
                        x, Pack
                    ), f"Selector tried to look inside script {x}"
                    for kid in x.kids:
                        if selector(kid):
                            next_results.append(kid)
                results = next_results
            all_results.extend(results)
        return MatchedSet(multipart_selectors, all_results)

    def __len__(self) -> int:
        count = 0
        for x in self.kids:
            match x:
                case Pack():
                    count += len(x)
                case Script():
                    count += 1
                case _:
                    raise ValueError(f"Unknown type: {type(x)}")
        return count

    @property
    def caption(self):
        return f"{super().caption}"

    def __format__(self, format_spec: RunnableFormat) -> str:
        format_spec = format_spec or "short"

        match format_spec:
            case "full":
                return pretty_print_kids(self.address, self.kids)
            case "line":
                return ": ".join(
                    [
                        self.name,
                        ", ".join(
                            f"{x:child}" for x in self.kids if x.is_visible
                        ),
                    ]
                )
            case "child":
                return f"{self.name}[{len(self)}]/"
            case "address":
                return f"{self.address}/"
            case "short":
                return self.caption + "/"
            case _:
                raise ValueError(f"Unknown format spec: {format_spec}")

    def __repr__(self) -> str:
        return self.__format__("line")

    def __str__(self) -> str:
        return self.__format__("short")

    def do_prerun(self, executor: BashPrefixExecutor):
        if self.pre_run:
            pr = executor.try_exec(
                cwd=self.pre_run.path.parent,
                path=self.pre_run.path,
                prefix=self.address,
            )
            if pr.returncode > 0:
                return False
        return True

    def run(self, executor: BashPrefixExecutor):
        for x in self.kids:
            x.run(executor)



from abc import ABC, abstractmethod
from dataclasses import dataclass
from pathlib import Path

from packe._exec.bash_exec_prefix import BashPrefixExecutor
from packe._scripts.types import RunnableFormat


@dataclass(eq=False)
class Runnable(ABC):
    pos: int | None
    name: str | None
    parent: "Runnable | None"

    def __hash__(self) -> int:
        return super().__hash__()

    def __eq__(self, value: object) -> bool:
        return super().__eq__(value)

    @property
    def is_visible(self):
        return self.name is not None

    @property
    def address(self):
        all: list[Runnable] = [*self.parents, self]
        prefix = "/".join([x.caption for x in all])
        return prefix

    @property
    def parents(self):
        all_parents: list[Runnable] = []
        last = self
        while last.parent:
            last = last.parent
            all_parents.append(last)
        all_parents.reverse()
        return all_parents

    @property
    def caption(self):
        parts = []
        if self.pos:
            parts.append(str(self.pos).zfill(2))
        parts.append(self.name)
        return ":".join(parts)

    @abstractmethod
    def __len__(self) -> int: ...

    @abstractmethod
    def run(self, executor: BashPrefixExecutor): ...

    @abstractmethod
    def __format__(self, format_spec: RunnableFormat) -> str: ...
    @abstractmethod
    def __repr__(self) -> str: ...

    @abstractmethod
    def __str__(self) -> str: ...



import os
from pathlib import Path
from shutil import which
from subprocess import STDOUT, Popen
from typing import Protocol

from termcolor import colored
from packe._root import package_root


class ExecInfo(Protocol):
    path: Path
    cwd: Path


class PrefixExecFailed(Exception, ExecInfo):
    path: Path
    cwd: Path

    def __init__(self, path: Path, cwd: Path):
        rel_path = path.relative_to(cwd)
        super().__init__(f"Failed to execute {rel_path} in {cwd}")
        self.cwd = cwd
        self.path = path


class BashPrefixExecutor:
    def __init__(self, before: Path | None = None):
        self.before = before

    def _build_env(self, prefix: str, path: Path):
        env_exec_dir = str(package_root / "packe" / "bash-exec")
        env_before = str(self.before if self.before else "")
        env_prefix = colored(f"[{prefix}] ", "cyan")
        env_target = str(path.absolute())
        env = {
            "PYRUN_EXEC_DIR": env_exec_dir,
            "PYRUN_BEFORE": env_before,
            "PYRUN_PREFIX": env_prefix,
            "PYRUN_TARGET": env_target,
            "PACKE_EXEC_DIR": env_exec_dir,
            "PACKE_BEFORE": env_before,
            "PACKE_PREFIX": env_prefix,
            "PACKE_TARGET": env_target,
            **os.environ,
        }
        return env

    def try_exec(
        self,
        path: Path,
        cwd: Path,
        prefix: str,
    ):
        exec_dir = package_root / "packe" / "bash-exec"
        exec_target = str(exec_dir / "exec.bash")
        bash_path = which("bash")
        if not bash_path:
            raise Exception("Failed to find bash")
        env = self._build_env(prefix, path)

        p = Popen(
            [bash_path, f"-c", f". {exec_target}", str(path)],
            shell=False,
            encoding="utf-8",
            env=env,
            cwd=cwd,
        )

        p.wait()
        return p

    def must_exec(
        self,
        path: Path,
        cwd: Path,
        prefix: str,
    ):
        p = self.try_exec(path, cwd, prefix)
        if p.returncode > 0:
            redline = colored(
                f"         ↑ FAILED AT {prefix} ↑         ",
                on_color="on_red",
                color="black",
            )
            print(redline, "\n")
            exit(1)



from pathlib import Path
import re


def try_parse_indexed(stem: str):
    # split on first dot or hyphen, so "01.name" and "01-name" both work
    parts = re.split(r"[.-]", stem, maxsplit=1)
    match parts:
        case [name]:
            return None
        case ["", name]:
            return None
        case ["_", name]:
            return (None, name)
        case [pos, name] if pos.isdigit():
            return (int(pos), name)
        case _:
            return None


def is_valid_script(path: Path):
    return path.suffix == ".sh" or path.suffix == ".bash"


def must_parse_indexed(stem: str):
    r = try_parse_indexed(stem)
    if r is None:
        raise Exception(f"Invalid indexed name: {stem}")
    return r
